/*
 * Copyright 2006 Google Inc.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.google.gwt.user.client.ui;

import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.Element;
import com.google.gwt.user.client.Event;

/**
 * A panel that stacks its children vertically, displaying only one at a time,
 * with a header for each that the user can click to display it.
 * 
 * <p>
 * <img class='gallery' src='StackPanel.png'/>
 * </p>
 * 
 * <h3>CSS Style Rules</h3>
 * <ul class='css'>
 * <li>.gwt-StackPanel { the panel itself }</li>
 * <li>.gwt-StackPanel .gwt-StackPanelItem { unselected items }</li>
 * <li>.gwt-StackPanel .gwt-StackPanelItem-selected { selected items }</li>
 * </ul>
 * 
 * <h3>Example</h3>
 * {@link com.google.gwt.examples.StackPanelExample code}
 */
public class StackPanel extends ComplexPanel {

  private Element body;
  private int visibleStack = -1;

  /**
   * Creates an empty stack panel.
   */
  public StackPanel() {
    Element table = DOM.createTable();
    setElement(table);

    body = DOM.createTBody();
    DOM.appendChild(table, body);
    DOM.setIntAttribute(table, "cellSpacing", 0);
    DOM.setIntAttribute(table, "cellPadding", 0);

    DOM.sinkEvents(table, Event.ONCLICK);
    setStyleName("gwt-StackPanel");
  }

  public boolean add(Widget w) {
    super.add(w);
    int index = getWidgetCount() - 1;

    Element tr = DOM.createTR();
    Element td = DOM.createTD();
    DOM.appendChild(body, tr);
    DOM.appendChild(tr, td);
    setStyleName(td, "gwt-StackPanelItem", true);
    DOM.setIntAttribute(td, "__index", index);
    DOM.setAttribute(td, "height", "1px");

    tr = DOM.createTR();
    td = DOM.createTD();
    DOM.appendChild(body, tr);
    DOM.appendChild(tr, td);
    DOM.appendChild(td, w.getElement());
    DOM.setAttribute(td, "height", "100%");
    DOM.setAttribute(td, "vAlign", "top");

    setStackVisible(index, false);
    if (visibleStack == -1)
      showStack(0);
    return true;
  }

  /**
   * Adds a widget to the panel, along with its header.
   * 
   * @param w the widget to be added
   * @param stackText the header text associated with this widget
   * @return <code>true</code> if successful
   */
  public boolean add(Widget w, String stackText) {
    return add(w, stackText, false);
  }

  /**
   * Adds a widget to the panel, along with its header, optionally interpreting
   * the header as HTML.
   * 
   * @param w the widget to be added
   * @param stackText the header text associated with this widget
   * @param asHTML <code>true</code> to treat the specified text as HTML
   * @return <code>true</code> if successful
   */
  public boolean add(Widget w, String stackText, boolean asHTML) {
    if (add(w)) {
      setStackText(getWidgetCount() - 1, stackText, asHTML);
      return true;
    }

    return false;
  }

  public void onBrowserEvent(Event event) {
    if (DOM.eventGetType(event) == Event.ONCLICK) {
      int index = getDividerIndex(DOM.eventGetTarget(event));
      if (index != -1)
        showStack(index);
    }
  }

  public boolean remove(Widget child) {
    if (child.getParent() != this)
      return false;

    // Remove the TR element containing the given child.
    int index = getWidgetIndex(child);
    Element tr = DOM.getChild(body, index);
    DOM.removeChild(body, tr);

    // Remove it from the widget hierarchy.
    return super.remove(child);
  }

  /**
   * Sets the text associated with a widget, by its index.
   * 
   * @param index the index of the widget whose text is to be set
   * @param text the text to be associated with it
   */
  public void setStackText(int index, String text) {
    setStackText(index, text, false);
  }

  /**
   * Sets the text associated with a widget, by its index.
   * 
   * @param index the index of the widget whose text is to be set
   * @param text the text to be associated with it
   * @param asHTML <code>true</code> to treat the specified text as HTML
   */
  public void setStackText(int index, String text, boolean asHTML) {
    if (index >= getWidgetCount())
      return;

    Element td = DOM.getChild(DOM.getChild(body, index * 2), 0);
    if (asHTML)
      DOM.setInnerHTML(td, text);
    else
      DOM.setInnerText(td, text);
  }

  /**
   * Shows the widget at the specified index.
   * 
   * @param index the widget index to be shown
   */
  public void showStack(int index) {
    if ((index >= getWidgetCount()) || (index == visibleStack))
      return;

    if (visibleStack >= 0)
      setStackVisible(visibleStack, false);

    visibleStack = index;
    setStackVisible(visibleStack, true);
  }

  private int getDividerIndex(Element elem) {
    while ((elem != null) && !DOM.compare(elem, getElement())) {
      String expando = DOM.getAttribute(elem, "__index");
      if (expando != null)
        return Integer.parseInt(expando);

      elem = DOM.getParent(elem);
    }

    return -1;
  }

  private void setStackVisible(int index, boolean visible) {
    // Get the first table row containing the widget's selector item.
    Element tr = DOM.getChild(body, (index * 2));
    if (tr == null)
      return;

    // Style the stack selector item.
    Element td = DOM.getFirstChild(tr);
    setStyleName(td, "gwt-StackPanelItem-selected", visible);

    // Show/hide the contained widget.
    tr = DOM.getChild(body, (index * 2) + 1);
    UIObject.setVisible(tr, visible);
    getWidget(index).setVisible(visible);
  }
}
